/*----------------------------------------------------------------------------* 
**  "comp2_5.cpp" using Romberg Integration
**
**  To build this in Visual C++ 6.0 perform the following steps:
**
**  1. Select New from File menu.
**
**  2. In the New dialog box select "Win32 Console Application" and enter 
**     a project name.  Then press the "Ok" button.
**
**  3. In the Win32 Application dialog box select "A simple application" 
**     then press the "Finish" button.  
**
**  4. Press "Ok" in the New Project Information dialog box.
**
**  5. Open the *.cpp file for entry of the program.  (We do not edit StdAfx.h
**     or StdAfx.cpp; VC++ requires these files.)
*/

#include "stdafx.h"
#include <stdio.h>
#include <math.h>

double integrate( double, double, int, double (f)(double) );
double Romberg( double, double, int, double (f)(double), double, int );
double Y( double );

/*----------------------------------------------------------------------------*  
**  "main" test of Romberg.
**
**  returns: 0
*/
int main()
{
    double a;                                   // approximate integral
    int    i;

    for( i = 2; i <= 8; i++ )
    {
	a = Romberg( 0.0, 3.0, i, Y, 1.0E-04, 1 );
	printf( "         n = %3d  area  = %12.9f\n", i, a );
    }

    return 0;
}

/*----------------------------------------------------------------------------*  
**  "Y" test function.
**
**  returns: value at time t
*/
double Y( double t )
{
    return fabs( sin( t ) ) / ( 3.0 * t + 1.0 );
}

/*----------------------------------------------------------------------------* 
**  "integrate" finds the definite integral of a specified function
**  using trapezoidal integration.
**
**  returns: nothing
*/
double integrate( double t_beg,                 // lower limit of integral
	          double delta,                 // period for integration
		  int    n,                     // number of segments > 0 
		  double (f)(double t) )        // integrand function
{
    double area;
    double t = t_beg;                           // initialize time
    double dt = delta / n;                      // initialize step size
    int    i;                                   // loop counter

    area = f( t_beg ) / 2.0;                    // initial value of integral 

    if( n > 1 )                                 // compute the n - 1 terms
    {                                           //   in the summation
	for( i = 1; i < n; i++ )
	{
	    t += dt;                            // increment time
	    area += f( t );                     // add next term to sum
	}
    }

    area += f( t_beg + delta ) / 2.0;           // add last term
    area *= dt;                                 // scale by delta-time

    return area;
}

/*----------------------------------------------------------------------------*  
**  "Romberg" finds the definite integral of a specified function
**  using the Romberg integration method.
**
**  returns: nothing
*/
double Romberg( double t_beg,                   // lower limit of integral
	        double delta,                   // period for integration
	        int    n_max,                   // maximum rows in R array 
	        double (f)(double t),           // integrand function
	        double err_max,                 // maximum result error 
	        int    test )                   // 0 for no test
{
    double error, fac;                          
    int    i = 1, m = 1;                        // initialize parameters 
    double r[21][21];

    r[i][1] = integrate( t_beg, delta, m, f );  // find array element r11

    do
    {
	i++;                                    // update row number
	m *= 2;                                 // update segment count 
	fac = 1.0;                              // init extrapolation factor

	// find first element in row
	r[i][1] = integrate( t_beg, delta, m, f );

	// use extrapolation to fine other elements of row
	for( int j = 2; j <= i; j++ )
	{
	    fac *= 4.0;
	    r[i][j] = ( fac * r[i][j-1] - r[i-1][j-1] ) / ( fac - 1.0 );
	}

	error = fabs( r[i][i] - r[i-1][i-1] );
    }
    while( i < n_max && error > err_max );        // terminating conditions

    if( test )
	printf( "Romberg: n = %3d  error = %12.9f\n", i, error );

    return r[i][i];                             // return integral estimate 
}